/*
 * Copyright © 2008 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * Authors:
 *    Eric Anholt <eric@anholt.net>
 *
 */

#include <linux/version.h>
#include <drmP.h>
#include <i915_drm.h>

#include "i915_reg.h"
#include "intel_drv.h"
#include "intel_bios.h"
#include "intel_ringbuffer.h"

#include "drm_emgd_private.h"

turbobuzy_t turbo;

int intel_pin_and_fence_fb_obj(struct drm_device *dev,
		struct drm_i915_gem_object *obj,
		struct intel_ring_buffer *pipelined)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	u32 alignment;
	int ret;

	switch (obj->tiling_mode) {
	case I915_TILING_NONE:
		if (IS_BROADWATER(dev) || IS_CRESTLINE(dev))
				alignment = 128 * 1024;
		else if (INTEL_INFO(dev)->gen >= 4)
				alignment = 4 * 1024;
		else
				alignment = 64 * 1024;
		break;
	case I915_TILING_X:
		/* pin() will align the object as required by fence */
		alignment = 0;
		break;
	case I915_TILING_Y:
		/* FIXME: Is this true? */
		DRM_ERROR("Y tiled not allowed for scan out buffers\n");
		return -EINVAL;
	default:
		BUG();
	}

	dev_priv->mm.interruptible = false;
	ret = i915_gem_object_pin_to_display_plane(obj, alignment, pipelined);
	if (ret)
			goto err_interruptible;

	/* Install a fence for tiled scan-out. Pre-i965 always needs a
	 * fence, whereas 965+ only requires a fence if using
	 * framebuffer compression.  For simplicity, we always install
	 * a fence as the cost is not that onerous.
	 */
	ret = i915_gem_object_get_fence(obj);
	if (ret)
		goto err_unpin;

	i915_gem_object_pin_fence(obj);

	dev_priv->mm.interruptible = true;
	return 0;

err_unpin:
	i915_gem_object_unpin(obj);
err_interruptible:
	dev_priv->mm.interruptible = true;
	return ret;
}

void notify_ring(struct drm_device *dev,
			struct intel_ring_buffer *ring)
{
	struct drm_i915_private *dev_priv = dev->dev_private;

#ifdef KERNEL_HAS_ATOMIC_PAGE_FLIP
	intel_atomic_notify_ring(dev, ring);
#endif

	if (ring->obj == NULL)
		return;

	trace_i915_gem_request_complete(ring, ring->get_seqno(ring, false));

	wake_up_all(&ring->irq_queue);
	if (i915_enable_hangcheck) {
		dev_priv->gpu_error.hangcheck_count = 0;
		mod_timer(&dev_priv->gpu_error.hangcheck_timer,
				round_jiffies_up(jiffies + DRM_I915_HANGCHECK_JIFFIES));
	}
}



/* NB: please notice the memset */
static void i915_get_extra_instdone(struct drm_device *dev,
		uint32_t *instdone)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	memset(instdone, 0, sizeof(*instdone) * I915_NUM_INSTDONE_REG);

	switch(INTEL_INFO(dev)->gen) {
		case 2:
		case 3:
			instdone[0] = I915_READ(INSTDONE);
			break;
		case 4:
		case 5:
		case 6:
			instdone[0] = I915_READ(INSTDONE_I965);
			instdone[1] = I915_READ(INSTDONE1);
			break;
		default:
			WARN_ONCE(1, "Unsupported platform\n");
		case 7:
			instdone[0] = I915_READ(GEN7_INSTDONE_1);
			instdone[1] = I915_READ(GEN7_SC_INSTDONE);
			instdone[2] = I915_READ(GEN7_SAMPLER_INSTDONE);
			instdone[3] = I915_READ(GEN7_ROW_INSTDONE);
			break;
	}
}

#ifdef CONFIG_DEBUG_FS
static struct drm_i915_error_object *
i915_error_object_create(struct drm_i915_private *dev_priv,
                         struct drm_i915_gem_object *src)
{
        struct drm_i915_error_object *dst;
        int i, count;
        u32 reloc_offset;

        if (src == NULL || src->pages == NULL)
                return NULL;

        count = src->base.size / PAGE_SIZE;

        dst = kmalloc(sizeof(*dst) + count * sizeof(u32 *), GFP_ATOMIC);
        if (dst == NULL)
                return NULL;

        reloc_offset = src->gtt_offset;
        for (i = 0; i < count; i++) {
                unsigned long flags;
                void *d;

                d = kmalloc(PAGE_SIZE, GFP_ATOMIC);
                if (d == NULL)
                        goto unwind;

                local_irq_save(flags);
                if (reloc_offset < dev_priv->mm.gtt_mappable_end &&
                    src->has_global_gtt_mapping) {
                        void __iomem *s;

                        /* Simply ignore tiling or any overlapping fence.
 			 * It's part of the error state, and this hopefully
                         * captures what the GPU read.
			 */

                        s = io_mapping_map_atomic_wc(dev_priv->mm.gtt_mapping,
                                                     reloc_offset);
                        memcpy_fromio(d, s, PAGE_SIZE);
                        io_mapping_unmap_atomic(s);
                } else {
			void *s;

                        struct page *page;
                        page = i915_gem_object_get_page(src, i);

                        drm_clflush_pages(&page, 1);

                        s = kmap_atomic(page);
                        memcpy(d, s, PAGE_SIZE);
                        kunmap_atomic(s);

                        drm_clflush_pages(&page, 1);
                }
                local_irq_restore(flags);

                dst->pages[i] = d;

                reloc_offset += PAGE_SIZE;
        }
        dst->page_count = count;
        dst->gtt_offset = src->gtt_offset;

        return dst;

unwind:
        while (i--)
                kfree(dst->pages[i]);
        kfree(dst);
        return NULL;
}

static void
i915_error_object_free(struct drm_i915_error_object *obj)
{
        int page;

        if (obj == NULL)
                return;

        for (page = 0; page < obj->page_count; page++)
                kfree(obj->pages[page]);

        kfree(obj);
}

void
i915_error_state_free(struct kref *error_ref)
{
        struct drm_i915_error_state *error = container_of(error_ref,
                                                          typeof(*error), ref);
        int i;

        for (i = 0; i < ARRAY_SIZE(error->ring); i++) {
                i915_error_object_free(error->ring[i].batchbuffer);
                i915_error_object_free(error->ring[i].ringbuffer);
                kfree(error->ring[i].requests);
        }

        kfree(error->active_bo);
        kfree(error->overlay);
        kfree(error->display);
        kfree(error);
}

static void capture_bo(struct drm_i915_error_buffer *err,
                       struct drm_i915_gem_object *obj)
{
        err->size = obj->base.size;
        err->name = obj->base.name;
        err->rseqno = obj->last_read_seqno;
        err->wseqno = obj->last_write_seqno;
        err->gtt_offset = obj->gtt_offset;
        err->read_domains = obj->base.read_domains;
        err->write_domain = obj->base.write_domain;
        err->fence_reg = obj->fence_reg;
        err->pinned = 0;
        if (obj->pin_count > 0)
                err->pinned = 1;
        if (obj->user_pin_count > 0)
                err->pinned = -1;
        err->tiling = obj->tiling_mode;
        err->dirty = obj->dirty;
        err->purgeable = obj->madv != I915_MADV_WILLNEED;
        err->ring = obj->ring ? obj->ring->id : -1;
        err->cache_level = obj->cache_level;
}

static u32 capture_active_bo(struct drm_i915_error_buffer *err,
                             int count, struct list_head *head)
{
        struct drm_i915_gem_object *obj;
        int i = 0;

        list_for_each_entry(obj, head, mm_list) {
                capture_bo(err++, obj);
                if (++i == count)
                        break;
        }

        return i;
}

static u32 capture_pinned_bo(struct drm_i915_error_buffer *err,
                             int count, struct list_head *head)
{
        struct drm_i915_gem_object *obj;
        int i = 0;

        list_for_each_entry(obj, head, gtt_list) {
                if (obj->pin_count == 0)
                        continue;

                capture_bo(err++, obj);
                if (++i == count)
                        break;
        }

        return i;
}

static void i915_gem_record_fences(struct drm_device *dev,
                                   struct drm_i915_error_state *error)
{
        struct drm_i915_private *dev_priv = dev->dev_private;
        int i;

        /* Fences */
        switch (INTEL_INFO(dev)->gen) {
        case 7:
        case 6:
                for (i = 0; i < 16; i++)
                        error->fence[i] = I915_READ64(FENCE_REG_SANDYBRIDGE_0 + (i * 8));
                break;
        case 5:
        case 4:
                for (i = 0; i < 16; i++)
                        error->fence[i] = I915_READ64(FENCE_REG_965_0 + (i * 8));
                break;
        case 3:
                if (IS_I945G(dev) || IS_I945GM(dev) || IS_G33(dev))
                        for (i = 0; i < 8; i++)
                                error->fence[i+8] = I915_READ(FENCE_REG_945_8 + (i * 4));
        case 2:
                for (i = 0; i < 8; i++)
                        error->fence[i] = I915_READ(FENCE_REG_830_0 + (i * 4));
                break;

        }
}

static struct drm_i915_error_object *
i915_error_first_batchbuffer(struct drm_i915_private *dev_priv,
                             struct intel_ring_buffer *ring)
{
        struct drm_i915_gem_object *obj;
        u32 seqno;

        if (!ring->get_seqno)
                return NULL;

        if (HAS_BROKEN_CS_TLB(dev_priv->dev)) {
                u32 acthd = I915_READ(ACTHD);

                if (WARN_ON(ring->id != RCS))
                        return NULL;

                obj = ring->private;
                if (acthd >= obj->gtt_offset &&
                    acthd < obj->gtt_offset + obj->base.size)
                        return i915_error_object_create(dev_priv, obj);
        }

        seqno = ring->get_seqno(ring, false);
        list_for_each_entry(obj, &dev_priv->mm.active_list, mm_list) {
                if (obj->ring != ring)
                        continue;

                if (i915_seqno_passed(seqno, obj->last_read_seqno))
                        continue;

                if ((obj->base.read_domains & I915_GEM_DOMAIN_COMMAND) == 0)
                        continue;

                /* We need to copy these to an anonymous buffer as the simplest
                 * method to avoid being overwritten by userspace.
                 */
                return i915_error_object_create(dev_priv, obj);
        }

        return NULL;
}

static void i915_record_ring_state(struct drm_device *dev,
                                   struct drm_i915_error_state *error,
                                   struct intel_ring_buffer *ring)
{
        struct drm_i915_private *dev_priv = dev->dev_private;

        if (INTEL_INFO(dev)->gen >= 6) {
                error->rc_psmi[ring->id] = I915_READ(ring->mmio_base + 0x50);
                error->fault_reg[ring->id] = I915_READ(RING_FAULT_REG(ring));
                error->semaphore_mboxes[ring->id][0]
                        = I915_READ(RING_SYNC_0(ring->mmio_base));
                error->semaphore_mboxes[ring->id][1]
                        = I915_READ(RING_SYNC_1(ring->mmio_base));
                error->semaphore_seqno[ring->id][0] = ring->sync_seqno[0];
                error->semaphore_seqno[ring->id][1] = ring->sync_seqno[1];
        }

        if (INTEL_INFO(dev)->gen >= 4) {
                error->faddr[ring->id] = I915_READ(RING_DMA_FADD(ring->mmio_base));
                error->ipeir[ring->id] = I915_READ(RING_IPEIR(ring->mmio_base));
                error->ipehr[ring->id] = I915_READ(RING_IPEHR(ring->mmio_base));
                error->instdone[ring->id] = I915_READ(RING_INSTDONE(ring->mmio_base));
                error->instps[ring->id] = I915_READ(RING_INSTPS(ring->mmio_base));
                if (ring->id == RCS)
                        error->bbaddr = I915_READ64(BB_ADDR);
        } else {
                error->faddr[ring->id] = I915_READ(DMA_FADD_I8XX);
                error->ipeir[ring->id] = I915_READ(IPEIR);
                error->ipehr[ring->id] = I915_READ(IPEHR);
                error->instdone[ring->id] = I915_READ(INSTDONE);
        }

        error->waiting[ring->id] = waitqueue_active(&ring->irq_queue);
        error->instpm[ring->id] = I915_READ(RING_INSTPM(ring->mmio_base));
        error->seqno[ring->id] = ring->get_seqno(ring, false);
        error->acthd[ring->id] = intel_ring_get_active_head(ring);
        error->head[ring->id] = I915_READ_HEAD(ring);
        error->tail[ring->id] = I915_READ_TAIL(ring);
        error->ctl[ring->id] = I915_READ_CTL(ring);

        error->cpu_ring_head[ring->id] = ring->head;
        error->cpu_ring_tail[ring->id] = ring->tail;
}

static void i915_gem_record_rings(struct drm_device *dev,
                                  struct drm_i915_error_state *error)
{
        struct drm_i915_private *dev_priv = dev->dev_private;
        struct intel_ring_buffer *ring;
        struct drm_i915_gem_request *request;
        int i, count;

        for_each_ring(ring, dev_priv, i) {
                i915_record_ring_state(dev, error, ring);

                error->ring[i].batchbuffer =
                        i915_error_first_batchbuffer(dev_priv, ring);

                error->ring[i].ringbuffer =
                        i915_error_object_create(dev_priv, ring->obj);

                count = 0;
                list_for_each_entry(request, &ring->request_list, list)
                        count++;

                error->ring[i].num_requests = count;
                error->ring[i].requests =
                        kmalloc(count*sizeof(struct drm_i915_error_request),
                                GFP_ATOMIC);
                if (error->ring[i].requests == NULL) {
                        error->ring[i].num_requests = 0;
                        continue;
                }

                count = 0;
                list_for_each_entry(request, &ring->request_list, list) {
                        struct drm_i915_error_request *erq;

                        erq = &error->ring[i].requests[count++];
                        erq->seqno = request->seqno;
                        erq->jiffies = request->emitted_jiffies;
                        erq->tail = request->tail;
                }
        }
}

/**
 * i915_capture_error_state - capture an error record for later analysis
 * @dev: drm device
 *
 * Should be called when an error is detected (either a hang or an error
 * interrupt) to capture error state from the time of the error.  Fills
 * out a structure which becomes available in debugfs for user level tools
 * to pick up.
 */

static void i915_capture_error_state(struct drm_device *dev)
{
        struct drm_i915_private *dev_priv = dev->dev_private;
        struct drm_i915_gem_object *obj;
        struct drm_i915_error_state *error;
        unsigned long flags;
        int i, pipe;

        spin_lock_irqsave(&dev_priv->error_lock, flags);
        error = dev_priv->first_error;
        spin_unlock_irqrestore(&dev_priv->error_lock, flags);
        if (error)
                return;

        /* Account for pipe specific data like PIPE*STAT */
        error = kzalloc(sizeof(*error), GFP_ATOMIC);
        if (!error) {
                DRM_DEBUG_DRIVER("out of memory, not capturing error state\n");
                return;
        }

        DRM_INFO("capturing error event; look for more information in /debug/dri/%d/i915_error_state\n",
                 dev->primary->index);

        kref_init(&error->ref);
        error->eir = I915_READ(EIR);
        error->pgtbl_er = I915_READ(PGTBL_ER);
        error->ccid = I915_READ(CCID);

        if (HAS_PCH_SPLIT(dev))
                error->ier = I915_READ(DEIER) | I915_READ(GTIER);
        else if (IS_VALLEYVIEW(dev))
                error->ier = I915_READ(GTIER) | I915_READ(VLV_IER);
        else if (IS_GEN2(dev))
                error->ier = I915_READ16(IER);
        else
                error->ier = I915_READ(IER);

        if (INTEL_INFO(dev)->gen >= 6)
                error->derrmr = I915_READ(DERRMR);

        if (IS_VALLEYVIEW(dev))
                error->forcewake = I915_READ(FORCEWAKE_VLV);
        else if (INTEL_INFO(dev)->gen >= 7)
                error->forcewake = I915_READ(FORCEWAKE_MT);
        else if (INTEL_INFO(dev)->gen == 6)
                error->forcewake = I915_READ(FORCEWAKE);

        for_each_pipe(pipe)
                error->pipestat[pipe] = I915_READ(PIPESTAT(pipe));

	if (INTEL_INFO(dev)->gen >= 6) {
                error->error = I915_READ(ERROR_GEN6);
                error->done_reg = I915_READ(DONE_REG);
        }

        if (INTEL_INFO(dev)->gen == 7)
                error->err_int = I915_READ(GEN7_ERR_INT);

        i915_get_extra_instdone(dev, error->extra_instdone);

        i915_gem_record_fences(dev, error);
        i915_gem_record_rings(dev, error);

        /* Record buffers on the active and pinned lists. */
        error->active_bo = NULL;
        error->pinned_bo = NULL;

        i = 0;
        list_for_each_entry(obj, &dev_priv->mm.active_list, mm_list)
                i++;
        error->active_bo_count = i;
        list_for_each_entry(obj, &dev_priv->mm.bound_list, gtt_list)
                if (obj->pin_count)
                        i++;
        error->pinned_bo_count = i - error->active_bo_count;

        error->active_bo = NULL;
        error->pinned_bo = NULL;
        if (i) {
                error->active_bo = kmalloc(sizeof(*error->active_bo)*i,
                                           GFP_ATOMIC);
                if (error->active_bo)
                        error->pinned_bo =
                                error->active_bo + error->active_bo_count;
        }

        if (error->active_bo)
                error->active_bo_count =
                        capture_active_bo(error->active_bo,
                                          error->active_bo_count,
                                          &dev_priv->mm.active_list);

        if (error->pinned_bo)
                error->pinned_bo_count =
                        capture_pinned_bo(error->pinned_bo,
                                          error->pinned_bo_count,
                                          &dev_priv->mm.bound_list);

	do_gettimeofday(&error->time);

        error->overlay = intel_overlay_capture_error_state(dev);
        error->display = intel_display_capture_error_state(dev);

        spin_lock_irqsave(&dev_priv->error_lock, flags);
        if (dev_priv->first_error == NULL) {
                dev_priv->first_error = error;
                error = NULL;
        }
        spin_unlock_irqrestore(&dev_priv->error_lock, flags);

        if (error)
                i915_error_state_free(&error->ref);
}

void i915_destroy_error_state(struct drm_device *dev)
{
        struct drm_i915_private *dev_priv = dev->dev_private;
        struct drm_i915_error_state *error;
        unsigned long flags;

        spin_lock_irqsave(&dev_priv->error_lock, flags);
        error = dev_priv->first_error;
        dev_priv->first_error = NULL;
        spin_unlock_irqrestore(&dev_priv->error_lock, flags);

        if (error)
                kref_put(&error->ref, i915_error_state_free);
}
#else
#define i915_capture_error_state(x)
#endif

static void i915_report_and_clear_eir(struct drm_device *dev)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	uint32_t instdone[I915_NUM_INSTDONE_REG];
	u32 eir = I915_READ(EIR);
	int pipe, i;

	if (!eir)
		return;

	printk(KERN_ERR "render error detected, EIR: 0x%08x\n",
	       eir);

	i915_get_extra_instdone(dev, instdone);

	if (IS_G4X(dev)) {
		if (eir & (GM45_ERROR_MEM_PRIV | GM45_ERROR_CP_PRIV)) {
			u32 ipeir = I915_READ(IPEIR_I965);

			printk(KERN_ERR "  IPEIR: 0x%08x\n",
			       I915_READ(IPEIR_I965));
			printk(KERN_ERR "  IPEHR: 0x%08x\n",
					I915_READ(IPEHR_I965));
			for (i = 0; i < ARRAY_SIZE(instdone); i++)
				pr_err("  INSTDONE_%d: 0x%08x\n", i, instdone[i]);
			printk(KERN_ERR "  INSTPS: 0x%08x\n",
			       I915_READ(INSTPS));
			printk(KERN_ERR "  ACTHD: 0x%08x\n",
			       I915_READ(ACTHD_I965));
			I915_WRITE(IPEIR_I965, ipeir);
			POSTING_READ(IPEIR_I965);
		}
		if (eir & GM45_ERROR_PAGE_TABLE) {
			u32 pgtbl_err = I915_READ(PGTBL_ER);
			printk(KERN_ERR "page table error\n");
			printk(KERN_ERR "  PGTBL_ER: 0x%08x\n",
			       pgtbl_err);
			I915_WRITE(PGTBL_ER, pgtbl_err);
			POSTING_READ(PGTBL_ER);
		}
	}

	if (!IS_GEN2(dev)) {
		if (eir & I915_ERROR_PAGE_TABLE) {
			u32 pgtbl_err = I915_READ(PGTBL_ER);
			printk(KERN_ERR "page table error\n");
			printk(KERN_ERR "  PGTBL_ER: 0x%08x\n",
			       pgtbl_err);
			I915_WRITE(PGTBL_ER, pgtbl_err);
			POSTING_READ(PGTBL_ER);
		}
	}

	if (eir & I915_ERROR_MEMORY_REFRESH) {
		printk(KERN_ERR "memory refresh error:\n");
		for_each_pipe(pipe)
			printk(KERN_ERR "pipe %c stat: 0x%08x\n",
			       pipe_name(pipe), I915_READ(PIPESTAT(pipe)));
		/* pipestat has already been acked */
	}
	if (eir & I915_ERROR_INSTRUCTION) {
		printk(KERN_ERR "instruction error\n");
		printk(KERN_ERR "  INSTPM: 0x%08x\n",
		       I915_READ(INSTPM));
		for (i = 0; i < ARRAY_SIZE(instdone); i++)
			pr_err("  INSTDONE_%d: 0x%08x\n", i, instdone[i]);
		if (INTEL_INFO(dev)->gen < 4) {
			u32 ipeir = I915_READ(IPEIR);

			printk(KERN_ERR "  IPEIR: 0x%08x\n",
			       I915_READ(IPEIR));
			printk(KERN_ERR "  IPEHR: 0x%08x\n",
			       I915_READ(IPEHR));
			printk(KERN_ERR "  ACTHD: 0x%08x\n",
			       I915_READ(ACTHD));
			I915_WRITE(IPEIR, ipeir);
			POSTING_READ(IPEIR);
		} else {
			u32 ipeir = I915_READ(IPEIR_I965);

			printk(KERN_ERR "  IPEIR: 0x%08x\n",
			       I915_READ(IPEIR_I965));
			printk(KERN_ERR "  IPEHR: 0x%08x\n",
			       I915_READ(IPEHR_I965));
			printk(KERN_ERR "  INSTPS: 0x%08x\n",
			       I915_READ(INSTPS));
			printk(KERN_ERR "  ACTHD: 0x%08x\n",
			       I915_READ(ACTHD_I965));
			I915_WRITE(IPEIR_I965, ipeir);
			POSTING_READ(IPEIR_I965);
		}
	}

	I915_WRITE(EIR, eir);
	POSTING_READ(EIR);
	eir = I915_READ(EIR);
	if (eir) {
		/*
		 * some errors might have become stuck,
		 * mask them.
		 */
		DRM_ERROR("EIR stuck: 0x%08x, masking\n", eir);
		I915_WRITE(EMR, I915_READ(EMR) | eir);
		I915_WRITE(IIR, I915_RENDER_COMMAND_PARSER_ERROR_INTERRUPT);
	}
}

static void intel_display_handle_reset(struct drm_device *dev)
{
	struct drm_crtc *crtc;
	/*
	 * Flips in the rings have been nuked by the reset,
	 * so complete all pending flips so that user space
	 * will get its events and not get stuck.
	 */
	 list_for_each_entry(crtc, &dev->mode_config.crtc_list, head) {
		emgd_crtc_t *emgd_crtc = container_of(crtc, emgd_crtc_t, base);

		if (emgd_crtc->flip_pending)
			atomic_inc(&emgd_crtc->unpin_work_count);

		intel_prepare_page_flip(dev, emgd_crtc->pipe);
		crtc_pageflip_handler(dev, emgd_crtc->pipe);
	}
}

static void i915_error_wake_up(struct drm_i915_private *dev_priv,
			       bool reset_completed)
{
	struct intel_ring_buffer *ring;
	int i;

	/*
	 * Notify all waiters for GPU completion events that reset state has
	 * been changed, and that they need to restart their wait after
	 * checking for potential errors (and bail out to drop locks if there is
	 * a gpu reset pending so that i915_error_work_func can acquire them).
	 */

	/* Wake up __wait_seqno, potentially holding dev->struct_mutex. */
	for_each_ring(ring, dev_priv, i)
		wake_up_all(&ring->irq_queue);

	/* Wake up intel_crtc_wait_for_pending_flips, holding crtc->mutex. */
		wake_up_all(&dev_priv->pending_flip_queue);

	/*
	 * Signal tasks blocked in i915_gem_wait_for_error that the pending
	 * reset state is cleared.
	 */
	if (reset_completed)
		wake_up_all(&dev_priv->gpu_error.reset_queue);
}

/**
 * i915_handle_error - handle an error interrupt
 * @dev: drm device
 *
 * Do some basic checking of regsiter state at error interrupt time and
 * dump it to the syslog.  Also call i915_capture_error_state() to make
 * sure we get a record and make it available in debugfs.  Fire a uevent
 * so userspace knows something bad happened (should trigger collection
 * of a ring dump etc.).
 */
void i915_handle_error(struct drm_device *dev, bool wedged)
{
	struct drm_i915_private *dev_priv = dev->dev_private;

	i915_capture_error_state(dev);
	i915_report_and_clear_eir(dev);

	if (wedged) {
		atomic_set(&dev_priv->gpu_error.reset_counter,
				I915_RESET_IN_PROGRESS_FLAG);
		atomic_set_mask(I915_RESET_IN_PROGRESS_FLAG,
				&dev_priv->gpu_error.reset_counter);

		/*
		 * Wakeup waiting processes so that the reset work function
		 * i915_error_work_func doesn't deadlock trying to grab various
		 * locks. By bumping the reset counter first, the woken
		 * processes will see a reset in progress and back off,
		 * releasing their locks and then wait for the reset completion.
		 * We must do this for _all_ gpu waiters that might hold locks
		 * that the reset work needs to acquire.
		 *
		 * Note: The wake_up serves as the required memory barrier to
		 * ensure that the waiters see the updated value of the reset
		 * counter atomic_t.

		 */
		i915_error_wake_up(dev_priv, false);

#ifdef KERNEL_HAS_ATOMIC_PAGE_FLIP
		intel_atomic_wedged(dev);
#endif
	}

	schedule_work(&dev_priv->gpu_error.work);
}

static u32
ring_last_seqno(struct intel_ring_buffer *ring)
{
	return list_entry(ring->request_list.prev,
			  struct drm_i915_gem_request, list)->seqno;
}

static bool i915_hangcheck_ring_idle(struct intel_ring_buffer *ring, bool *err)
{
	if (list_empty(&ring->request_list) ||
	    i915_seqno_passed(ring->get_seqno(ring, false),
				          ring_last_seqno(ring))) {
		/* Issue a wake-up to catch stuck h/w. */
		if (waitqueue_active(&ring->irq_queue)) {
			DRM_ERROR("Hangcheck timer elapsed... %s idle\n",
					ring->name);
			wake_up_all(&ring->irq_queue);
			*err = true;
		}
		return true;
	}
	return false;
}

static bool kick_ring(struct intel_ring_buffer *ring)
{
	struct drm_device *dev = ring->dev;
	struct drm_i915_private *dev_priv = dev->dev_private;
	u32 tmp = I915_READ_CTL(ring);
	if (tmp & RING_WAIT) {
		DRM_ERROR("Kicking stuck wait on %s\n",
			  ring->name);
		I915_WRITE_CTL(ring, tmp);
		return true;
	}
	if (IS_GEN6(dev) &&
	    (tmp & RING_WAIT_SEMAPHORE)) {
		DRM_ERROR("Kicking stuck semaphore on %s\n",
			  ring->name);
		I915_WRITE_CTL(ring, tmp);
		return true;
	}
	return false;
}

static bool i915_hangcheck_hung(struct drm_device *dev)
{
	drm_i915_private_t *dev_priv = dev->dev_private;

	if (dev_priv->gpu_error.hangcheck_count++ > 1) {
		bool hung = true;

		DRM_ERROR("Hangcheck timer elapsed... GPU hung\n");
		i915_handle_error(dev, true);

		if (!IS_GEN2(dev)) {
			struct intel_ring_buffer *ring;
			int i;

			/* Is the chip hanging on a WAIT_FOR_EVENT?
			 * If so we can simply poke the RB_WAIT bit
			 * and break the hang. This should work on
			 * all but the second generation chipsets.
			 */
			for_each_ring(ring, dev_priv, i)
				hung &= !kick_ring(ring);
		}

		return hung;
	}

	return false;
}

/**
 * This is called when the chip hasn't reported back with completed
 * batchbuffers in a long time. The first time this is called we simply record
 * ACTHD. If ACTHD hasn't changed by the time the hangcheck timer elapses
 * again, we assume the chip is wedged and try to fix it.
 */
/* #define RINGBUFFER_DEBUG */
void i915_hangcheck_elapsed(unsigned long data)
{
	struct drm_device *dev = (struct drm_device *)data;
	drm_i915_private_t *dev_priv = dev->dev_private;
	uint32_t acthd[I915_NUM_RINGS], instdone[I915_NUM_INSTDONE_REG];
	struct intel_ring_buffer *ring;
	bool err = false, idle;
	int i;
#ifdef RINGBUFFER_DEBUG
	uint32_t tmp, start, head, tail;
	unsigned char * mmioptr;
#endif
	if (!i915_enable_hangcheck)
		return;

	memset(acthd, 0, sizeof(acthd));
	idle = true;
	for_each_ring(ring, dev_priv, i) {
		idle &= i915_hangcheck_ring_idle(ring, &err);
		acthd[i] = intel_ring_get_active_head(ring);
	}

	/* If all work is done then ACTHD clearly hasn't advanced. */
	if (idle) {
		if (err) {
			if (i915_hangcheck_hung(dev))
				return;

			goto repeat;
		}

		dev_priv->gpu_error.hangcheck_count = 0;
		return;
	}

	i915_get_extra_instdone(dev, instdone);

	if (memcmp(dev_priv->gpu_error.last_acthd, acthd, sizeof(acthd)) == 0 &&
			memcmp(dev_priv->gpu_error.prev_instdone, instdone, sizeof(instdone)) == 0) {
#ifdef RINGBUFFER_DEBUG
			{
				/*
				0x020XX --> Render
				0x120XX --> Video
				0x220XX --> Blitter
				*/

				mmioptr = dev_priv->regs;
				tmp = *(unsigned long *)(mmioptr + 0x1820b0);
				DRM_ERROR("EIR for MASTER_ERROR = 0x%08x", tmp);

				DRM_ERROR("Dump for Blitter RingBuffer\n");
				DRM_ERROR("---------------------------\n");
				tmp = I915_READ(0x2203C);//RING_BUFFER_CTL
				DRM_ERROR("     -----> RING_BUFFER_CTL   = 0x%08x\n", tmp);
				start = I915_READ(0x22038);//RING_BUFFER_START
				DRM_ERROR("     -----> RING_BUFFER_START = 0x%08x\n", start);
				head = I915_READ(0x22034);//RING_BUFFER_HEAD
				DRM_ERROR("     -----> RING_BUFFER_HEAD = 0x%08x\n", head);
				head &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN HEAD = 0x%08x\n", head);
				tail = I915_READ(0x22030);//RING_BUFFER_TAIL
				DRM_ERROR("     -----> RING_BUFFER_TAIL = 0x%08x\n", tail);
				tail &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN TAIL = 0x%08x\n", tail);
				tmp = I915_READ(0x2206C);//INSTDONE
				DRM_ERROR("     ----->       INSTDONE   = 0x%08x\n", tmp);
				tmp = I915_READ(0x22064);//IPEIR
				DRM_ERROR("     ----->       IPEIR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x22068);//IPEHR
				DRM_ERROR("     ----->       IPEHR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x22074);//ACTHD
				DRM_ERROR("     ----->       ACTHD      = 0x%08x\n", tmp);
				tmp = I915_READ(0x220B0);//EIR
				DRM_ERROR("     ----->       EIR        = 0x%08x\n", tmp);
				tmp = I915_READ(0x22070);//INSTPS
				DRM_ERROR("     ----->       INSTPS     = 0x%08x\n", tmp);
				tmp = I915_READ(0x22078);//DMA_FADD
				DRM_ERROR("     ----->       DMA_FADD   = 0x%08x\n", tmp);

				DRM_ERROR("Dump for Render RingBuffer\n");
				DRM_ERROR("---------------------------\n");
				tmp = I915_READ(0x0203C);//RING_BUFFER_CTL
				DRM_ERROR("     -----> RING_BUFFER_CTL   = 0x%08x\n", tmp);
				start = I915_READ(0x02038);//RING_BUFFER_START
				DRM_ERROR("     -----> RING_BUFFER_START = 0x%08x\n", start);
				head = I915_READ(0x02034);//RING_BUFFER_HEAD
				DRM_ERROR("     -----> RING_BUFFER_HEAD = 0x%08x\n", head);
				head &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN HEAD = 0x%08x\n", head);
				tail = I915_READ(0x02030);//RING_BUFFER_TAIL
				DRM_ERROR("     -----> RING_BUFFER_TAIL = 0x%08x\n", tail);
				tail &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN TAIL = 0x%08x\n", tail);
				tmp = I915_READ(0x0206C);//INSTDONE
				DRM_ERROR("     ----->       INSTDONE   = 0x%08x\n", tmp);
				tmp = I915_READ(0x02064);//IPEIR
				DRM_ERROR("     ----->       IPEIR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x02068);//IPEHR
				DRM_ERROR("     ----->       IPEHR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x02074);//ACTHD
				DRM_ERROR("     ----->       ACTHD      = 0x%08x\n", tmp);
				tmp = I915_READ(0x020B0);//EIR
				DRM_ERROR("     ----->       EIR        = 0x%08x\n", tmp);
				tmp = *(unsigned long *)(mmioptr + 0x020b0);
				DRM_ERROR("     ----->       EIR        = 0x%08x\n", tmp);
				tmp = I915_READ(0x02070);//INSTPS
				DRM_ERROR("     ----->       INSTPS     = 0x%08x\n", tmp);
				tmp = I915_READ(0x02078);//DMA_FADD
				DRM_ERROR("     ----->       DMA_FADD   = 0x%08x\n", tmp);


				DRM_ERROR("Dump for Video RingBuffer\n");
				DRM_ERROR("---------------------------\n");
				tmp = I915_READ(0x1203C);//RING_BUFFER_CTL
				DRM_ERROR("     -----> RING_BUFFER_CTL   = 0x%08x\n", tmp);
				start = I915_READ(0x12038);//RING_BUFFER_START
				DRM_ERROR("     -----> RING_BUFFER_START = 0x%08x\n", start);
				head = I915_READ(0x12034);//RING_BUFFER_HEAD
				DRM_ERROR("     -----> RING_BUFFER_HEAD = 0x%08x\n", head);
				head &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN HEAD = 0x%08x\n", head);
				tail = I915_READ(0x12030);//RING_BUFFER_TAIL
				DRM_ERROR("     -----> RING_BUFFER_TAIL = 0x%08x\n", tail);
				tail &= 0x000FFFFF;
				DRM_ERROR("     ----->       CLEAN TAIL = 0x%08x\n", tail);
				tmp = I915_READ(0x1206C);//INSTDONE
				DRM_ERROR("     ----->       INSTDONE   = 0x%08x\n", tmp);
				tmp = I915_READ(0x12064);//IPEIR
				DRM_ERROR("     ----->       IPEIR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x12068);//IPEHR
				DRM_ERROR("     ----->       IPEHR      = 0x%08x\n", tmp);
				tmp = I915_READ(0x12074);//ACTHD
				DRM_ERROR("     ----->       ACTHD      = 0x%08x\n", tmp);
				tmp = I915_READ(0x120B0);//EIR
				DRM_ERROR("     ----->       EIR        = 0x%08x\n", tmp);
				tmp = I915_READ(0x12070);//INSTPS
				DRM_ERROR("     ----->       INSTPS     = 0x%08x\n", tmp);
				tmp = I915_READ(0x12078);//DMA_FADD
				DRM_ERROR("     ----->       DMA_FADD   = 0x%08x\n", tmp);

			}
#endif

			if (i915_hangcheck_hung(dev))
				return;
	} else {
		dev_priv->gpu_error.hangcheck_count = 0;

		memcpy(dev_priv->gpu_error.last_acthd, acthd, sizeof(acthd));
		memcpy(dev_priv->gpu_error.prev_instdone, instdone, sizeof(instdone));
	}

repeat:
	/* Reset timer case chip hangs without another request being added */
	mod_timer(&dev_priv->gpu_error.hangcheck_timer,
			round_jiffies_up(jiffies + DRM_I915_HANGCHECK_JIFFIES));
}

/**
 * i915_error_work_func - do process context error handling work
 * @work: work struct
 *
 * Fire an error uevent so userspace can see that a hang or error
 * was detected.
 */
void i915_error_work_func(struct work_struct *work)
{
	struct i915_gpu_error *error = container_of(work, struct i915_gpu_error,
			work);
	drm_i915_private_t *dev_priv = container_of(error, drm_i915_private_t,
			gpu_error);
	struct drm_device *dev = dev_priv->dev;
	char *error_event[] = { "ERROR=1", NULL };
	char *reset_event[] = { "RESET=1", NULL };
	char *reset_done_event[] = { "ERROR=0", NULL };
	int ret;

	kobject_uevent_env(&dev->primary->kdev.kobj, KOBJ_CHANGE, error_event);

	if (i915_reset_in_progress(error)) {
		DRM_DEBUG_DRIVER("resetting chip\n");

		/* We aren't calling an i915_reset, however there are things we need to
		 * do that are in there, so they are here.
		 */
		dev_priv->gpu_error.stop_rings = 0;

		kobject_uevent_env(&dev->primary->kdev.kobj, KOBJ_CHANGE, reset_event);
		
		ret = i915_reset(dev);
		intel_display_handle_reset(dev);
		
		if (ret == 0) {
			/*
			 * After all the gem state is reset, increment the reset
			 * counter and wake up everyone waiting for the reset to
			 * complete.
			 *
			 * Since unlock operations are a one-sided barrier only,
			 * we need to insert a barrier here to order any seqno
			 * updates before
			 * the counter increment.
			 */
			smp_mb__before_atomic_inc();
			atomic_inc(&dev_priv->gpu_error.reset_counter);

			kobject_uevent_env(&dev->primary->kdev.kobj,
						KOBJ_CHANGE, reset_done_event);
		} else {
			atomic_set(&error->reset_counter, I915_WEDGED);
		}

		/*
		 * Note: The wake_up also serves as a memory barrier so that
		 *  waiters see the update value of the reset counter atomic_t.
		 */
		i915_error_wake_up(dev_priv, true);
	}
}

void intel_unpin_fb_obj(struct drm_i915_gem_object *obj)
{
	i915_gem_object_unpin_fence(obj);
	i915_gem_object_unpin(obj);
}


void vlv_force_wake_get(struct drm_i915_private *dev_priv)
{
	int count;

	count = 0;

	/* Already awake? */
	if ((I915_READ(0x130094) & 0xa1) == 0xa1)
		return;

	I915_WRITE_NOTRACE(FORCEWAKE_VLV, 0xffffffff);
	POSTING_READ(FORCEWAKE_VLV);

	count = 0;
	while (count++ < 50 && (I915_READ_NOTRACE(FORCEWAKE_ACK_VLV) & 1) == 0)
		udelay(10);
}

void vlv_force_wake_put(struct drm_i915_private *dev_priv)
{
	I915_WRITE_NOTRACE(FORCEWAKE_VLV, 0xffff0000);
	/* FIXME: confirm VLV behavior with Punit folks */
	POSTING_READ(FORCEWAKE_VLV);
}


static int gen6_do_reset(struct drm_device *dev)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	int	ret;
	unsigned long irqflags;

	/* Hold gt_lock across reset to prevent any register access
	 * with forcewake not set correctly
	 */
	spin_lock_irqsave(&dev_priv->gt_lock, irqflags);

	/* Reset the chip */

	/* GEN6_GDRST is not in the gt power well, no need to check
	 * for fifo space for the write or forcewake the chip for
	 * the read
	 */
	I915_WRITE_NOTRACE(GEN6_GDRST, GEN6_GRDOM_FULL);

	/* Spin waiting for the device to ack the reset request */
	ret = wait_for((I915_READ_NOTRACE(GEN6_GDRST) & GEN6_GRDOM_FULL) == 0, 500);

	/* If reset with a user forcewake, try to restore, otherwise turn it off */
	if (dev_priv->forcewake_count)
		vlv_force_wake_get(dev_priv);
	else
		vlv_force_wake_put(dev_priv);

	/* Restore fifo count */
	dev_priv->gt_fifo_count = I915_READ_NOTRACE(GT_FIFO_FREE_ENTRIES);

	spin_unlock_irqrestore(&dev_priv->gt_lock, irqflags);
	return ret;
}

int intel_gpu_reset(struct drm_device *dev)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	int ret = -ENODEV;

	switch (INTEL_INFO(dev)->gen) {
	case 7:
	case 6:
		ret = gen6_do_reset(dev);
		break;
	}

	/* Also reset the gpu hangman. */
	if (dev_priv->gpu_error.stop_rings) {
		DRM_DEBUG("Simulated gpu hang, resetting stop_rings\n");
		dev_priv->gpu_error.stop_rings = 0;
		if (ret == -ENODEV) {
			DRM_ERROR("Reset not implemented, but ignoring "
				  "error for simulated gpu hangs\n");
			ret = 0;
		}
	}

	return ret;
}


/**
 * i915_reset - reset chip after a hang
 * @dev: drm device to reset
 *
 * Reset the chip.  Useful if a hang is detected. Returns zero on successful
 * reset or otherwise an error code.
 *
 * Procedure is fairly simple:
 *   - reset the chip using the reset reg
 *   - re-init context state
 *   - re-init hardware status page
 *   - re-init ring buffer
 *   - re-init interrupt state
 *   - re-init display
 */
int i915_reset(struct drm_device *dev)
{
	drm_i915_private_t *dev_priv = dev->dev_private;
	bool simulated;
	int ret;

	if (!i915_try_reset)
		return 0;

	mutex_lock(&dev->struct_mutex);

	i915_gem_reset(dev);

	simulated = dev_priv->gpu_error.stop_rings != 0;

	if (!simulated && get_seconds() - dev_priv->gpu_error.last_reset < 5) {
		DRM_ERROR("GPU hanging too fast, declaring wedged!\n");
		ret = -ENODEV;
	} else {
		ret = intel_gpu_reset(dev);

		/* Also reset the gpu hangman. */
		if (simulated) {
			DRM_INFO("Simulated gpu hang, resetting stop_rings\n");
			dev_priv->gpu_error.stop_rings = 0;
			if (ret == -ENODEV) {
				DRM_ERROR("Reset not implemented, but ignoring "
					  "error for simulated gpu hangs\n");
				ret = 0;
			}
		} else
			dev_priv->gpu_error.last_reset = get_seconds();
	}
	if (ret) {
		DRM_ERROR("Failed to reset chip.\n");
		mutex_unlock(&dev->struct_mutex);
		return ret;
	}

	/* Ok, now get things going again... */

	/*
	 * Everything depends on having the GTT running, so we need to start
	 * there.  Fortunately we don't need to do this unless we reset the
	 * chip at a PCI level.
	 *
	 * Next we need to restore the context, but we don't use those
	 * yet either...
	 *
	 * Ring buffer needs to be re-initialized in the KMS case, or if X
	 * was running at the time of the reset (i.e. we weren't VT
	 * switched away).
	 */
	if (drm_core_check_feature(dev, DRIVER_MODESET) ||
			!dev_priv->mm.suspended) {
		struct intel_ring_buffer *ring;
		int i;

		dev_priv->mm.suspended = 0;

		i915_gem_init_swizzling(dev);

		for_each_ring(ring, dev_priv, i)
			ring->init(ring);

		i915_gem_context_init(dev);
		
		/* Enable after with ppgtt support */
		/*
		if (dev_priv->mm.aliasing_ppgtt) {
			ret = dev_priv->mm.aliasing_ppgtt->enable(dev);
			if (ret)
				i915_gem_cleanup_aliasing_ppgtt(dev);
		}
		*/
		/*
		 * It would make sense to re-init all the other hw state, at
		 * least the rps/rc6/emon init done within modeset_init_hw. For
		 * some unknown reason, this blows up my ilk, so don't.
		 */

		mutex_unlock(&dev->struct_mutex);

		drm_irq_uninstall(dev);
		drm_irq_install(dev);
		/* EMGD doesn't have hotplug support yet 
 		 * Uncomment below only after add hotplug
 		 */
		/*	
 		intel_hpd_init(dev); 
		*/
	} else {
		mutex_unlock(&dev->struct_mutex);
	}

	return 0;
}

void vlv_queue_rps_work(struct drm_i915_private *dev_priv,
                                u32 pm_iir)
{
	unsigned long flags;
	/* IIR bits should never already be set because IMR should
	prevent an interrupt from being shown in IIR. The warning
	displays a case where we've unsafely cleared
	dev_priv->rps.pm_iir. Although missing an interrupt of the same
	type is not a problem, it displays a problem in the logic.

	The mask bit in IMR is cleared by dev_priv->rps.work.*/

	spin_lock_irqsave(&dev_priv->rps.lock, flags);
	dev_priv->rps.pm_iir |= pm_iir;

	spin_unlock_irqrestore(&dev_priv->rps.lock, flags);
	queue_work(dev_priv->wq, &dev_priv->rps.work);
}

void vlv_pm_rps_work(struct work_struct *work){

	static int turbo_flag = 0, turbo_down_cnt = 0;
	unsigned long long constant = 10000000000, render = 0, media = 0;
	u32 turbo_new_freq = 0;

	drm_i915_private_t *dev_priv = container_of(work, drm_i915_private_t,
						   rps.work);
	u32 pm_iir;

	spin_lock_irq(&dev_priv->rps.lock);
	pm_iir = dev_priv->rps.pm_iir;
	dev_priv->rps.pm_iir = 0;
	spin_unlock_irq(&dev_priv->rps.lock);

	if (dev_priv->rps.max_lock || dev_priv->mem_freq == 0)
		return;

	mutex_lock(&dev_priv->rps.hw_lock);

	/* The calculation throttling equation:
	Rc0 time = RC0 cnt delta / (CZ clock freq / 1000) ms
	Evaluate Interval = IOSF(0xCE) delta / 100000 ms
	RC0 residency = RC0 time * 100 / Evaluation Interval

	RC0 residency = ( RC0 cnt delta / OSF(0xCE) delta ) *
	( 10^10 / dev->mem_freq)
	*/

	do_div(constant,dev_priv->mem_freq);

	if (turbo_flag == 0){
		turbo.render_prev=I915_READ(RENDER_RC0_COUNTER);
		turbo.media_prev=I915_READ(MEDIA_RC0_COUNTER);
		valleyview_punit_read(dev_priv, GR_PORT_LSB, &(turbo.punit_timestamp_prev));
		turbo_flag++;
	}else{
		turbo.render_cur=I915_READ(RENDER_RC0_COUNTER);
		turbo.media_cur=I915_READ(MEDIA_RC0_COUNTER);
		valleyview_punit_read(dev_priv, GR_PORT_LSB, &(turbo.punit_timestamp_cur));

		turbo.render_prev = (turbo.render_prev > turbo.render_cur)?
				     0xFFFFFFFF - turbo.render_prev + turbo.render_cur:
				     turbo.render_cur - turbo.render_prev;

		turbo.media_prev = (turbo.media_prev > turbo.media_cur)?
				    0xFFFFFFFF - turbo.media_prev + turbo.media_cur:
				    turbo.media_cur - turbo.media_prev;

		render = turbo.render_prev * constant;
		media = turbo.media_prev * constant;

		turbo.punit_timestamp_prev = ( turbo.punit_timestamp_prev > turbo.punit_timestamp_cur)?
					0xFFFFFFFF - turbo.punit_timestamp_prev + turbo.punit_timestamp_cur:
					turbo.punit_timestamp_cur - turbo.punit_timestamp_prev;


		if (turbo.punit_timestamp_prev == 0)
			turbo.punit_timestamp_prev=1;

		do_div(render, turbo.punit_timestamp_prev);
		do_div(media, turbo.punit_timestamp_prev);

		/*
		 * Media and render in this case refers to residency computation to 
		 * decide whether we need to boost up the GPU frequency or not. 
		 * 90000 render residency basically refers to the maximum threshold
		 * value we baceuse the gives a very huge number. 
		 *
		 * However, in certain cases, eg,video encoding/decoding, there 
		 * is some contribution to the media residency as well. Hence, we 
		 * have to take this into account as well. The value threshold is
		 * set to really low becase when mpeg media recidencies much than
		 * than h264 media residencies. 
		 */

		if (render > 900000 || media > 1000) {
			turbo_down_cnt = 0;
			turbo_new_freq = dev_priv->rps.cur_delay + 3;
			if (!(turbo_new_freq > dev_priv->rps.max_delay))
				valleyview_set_rps(dev_priv->dev, turbo_new_freq);
			else
				valleyview_set_rps(dev_priv->dev, dev_priv->rps.max_delay);
			
		}else{
			if(++turbo_down_cnt%5 == 0) { /* used to control rate of decrease */
				turbo_new_freq = dev_priv->rps.cur_delay - 1;
				if (!(turbo_new_freq < dev_priv->rps.rpe_delay))
					valleyview_set_rps(dev_priv->dev, turbo_new_freq);
				else
					valleyview_set_rps(dev_priv->dev, dev_priv->rps.rpe_delay);
			}
		}

		turbo.render_prev = turbo.render_cur;
		turbo.media_prev = turbo.media_cur;
		turbo.punit_timestamp_prev = turbo.punit_timestamp_cur;

		/* On VLV, when we enter RC6 we may not be at the minimum
		 * voltage level, so arm a timer to check.  It should only
		 * fire when there's activity or once after we've entered
		 * RC6, and then won't be re-armed until the next RPS interrupt.
		 */
		mod_delayed_work(dev_priv->wq, &dev_priv->rps.vlv_work,
			msecs_to_jiffies(100));

	}
	mutex_unlock(&dev_priv->rps.hw_lock);
}

