/* -*- pse-c -*-
 *-----------------------------------------------------------------------------
 * Filename: init_snb.c
 *-----------------------------------------------------------------------------
 * Copyright (c) 2002-2013, Intel Corporation.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files \(the "Software"\), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * 
 *-----------------------------------------------------------------------------
 * Description:
 *  
 *-----------------------------------------------------------------------------
 */

#include <io.h>
#include <pci.h>
#include <memmap.h>
#include <memory.h>

#include <igd_core_structs.h>

#include <asm/cacheflush.h>
#include <gn6/regs.h>
#include <gn6/context.h>

#include "../cmn/init_dispatch.h"

/*!
 * @addtogroup core_group
 * @{
 */

static int snb_bus_master_enable(platform_context_snb_t *platform_context);
static int full_config_vga_snb(igd_context_t *context,
	init_dispatch_t *dispatch);
static int get_stolen_mem_snb(igd_context_t *context, unsigned long *pages);
extern int get_param_snb(igd_context_t *context, unsigned long id,
	unsigned long *value);
int gtt_init_snb(igd_context_t *context);

/*!
 * 
 * @param context
 * @param dispatch
 * @param platform_context
 * 
 * @return -IGD_ERROR_NODEV on failure
 * @return 0 on success
 */
int full_query_bridge_snb(
	igd_context_t *context,
	init_dispatch_t *dispatch,
	platform_context_snb_t *platform_context)
{
	os_pci_dev_t pdev = (os_pci_dev_t)0;

	EMGD_TRACE_ENTER;

	pdev = OS_PCI_FIND_DEVICE(
			PCI_VENDOR_ID_INTEL,
			PCI_DEVICE_ID_BRIDGE_SNB,
			0xFFFF, /* Scan the whole PCI bus */
			0,
			0,
			(os_pci_dev_t)0);

	if(!pdev) {
		EMGD_ERROR_EXIT("Bridge device NOT found.");
		return -IGD_ERROR_NODEV;
	}

	EMGD_DEBUG("Bridge device found: 0x%x", PCI_DEVICE_ID_BRIDGE_SNB);
	platform_context->dramc_dev.pcidev = pdev;
	platform_context->dramc_dev.did = PCI_DEVICE_ID_BRIDGE_SNB;
	context->device_context.bid = PCI_DEVICE_ID_BRIDGE_SNB;

	EMGD_TRACE_EXIT;
	return 0;
}

/*!
 * 
 * @param context
 * @param dispatch
 * @param platform_context
 * 
 * @return 0 on success
 * @return 1 on failure
 */
int full_config_snb(igd_context_t *context,
	init_dispatch_t *dispatch)
{
	platform_context_snb_t *platform_context;
	unsigned long reserved_mem;
	int ret;

	EMGD_TRACE_ENTER;

	platform_context = (platform_context_snb_t *)context->platform_context;

	/* Bus master enabling for Gen4 platform */
	ret = snb_bus_master_enable(platform_context);
	if(ret) {
		EMGD_ERROR("Enabling bus master");
	}

	/* Config VGA */
	ret = full_config_vga_snb(context, dispatch);
	if(ret) {
		EMGD_ERROR_EXIT("Config VGA Failed");
		return ret;
	}

	/* get stolen memory from the HW / BIOS registers */
	get_stolen_mem_snb(context, &context->device_context.reserved_mem);

	/* Get mem reservation param from firmware if it exists - overrides the HW regs*/
	if(!get_param_snb(context, IGD_PARAM_MEM_RESERVATION, &reserved_mem)) {
		context->device_context.reserved_mem = reserved_mem;
	}

	gtt_init_snb(context);

	/* Set the chipset revision id */
	context->device_context.rid = platform_context->igfx_dev.rid;

	EMGD_TRACE_EXIT;
	return 0;
}

/*!
 * 
 * @param platform_context
 * 
 * @return -1 on failure
 * @return 0 on success
 */
static int snb_bus_master_enable(platform_context_snb_t *platform_context){
	int ret;
	unsigned char tmp;

	EMGD_TRACE_ENTER;

	ret = OS_PCI_READ_CONFIG_8(platform_context->igfx_dev.pcidev0, 0x4, &tmp);
	if(ret) {
		EMGD_ERROR_EXIT("PCI read of bus master");
		return -1;
	}

	/* Get Bit 2, 1, and 0 and see if it is == 1
	 * all 3 bits has to be enabled. This is to enable register read/write
	 * in the case of a PCI card being added */
	if((tmp & 0x7) != 0x7 ){

		tmp |= 0x7;

		ret = OS_PCI_WRITE_CONFIG_8(platform_context->igfx_dev.pcidev0, 0x4, tmp);
		if(ret) {
			EMGD_ERROR_EXIT("Error: PCI write of bus master");
			return -1;
		}
	}

	EMGD_TRACE_EXIT;
	return 0;
}

/*!
 * 
 * @param context
 * @param dispatch
 * 
 * @return -IGD_ERROR_NODEV on failure
 * @return 0 on success
 */
static int full_config_vga_snb(igd_context_t *context,
	init_dispatch_t *dispatch)
{
	uint32_t tmp32;
	unsigned long gttmmadr_size;
	unsigned char tmp = 0;
	int size  = 0;
	int ret = 0;

	platform_context_snb_t *platform_context =
		(platform_context_snb_t *)context->platform_context;

	EMGD_TRACE_ENTER;

	/* Read SNB MSAC */
	ret=OS_PCI_READ_CONFIG_8(platform_context->igfx_dev.pcidev0,
		INTEL_OFFSET_VGA_MSAC_SNB, (void*)&tmp);

	/* SNB GTTMMADR is 4MB */
	gttmmadr_size = 4*1024*1024;

	if(!ret){
		switch (tmp & 0x6) {
		case 0x2:
			size = 256*1024*1024;
			context->device_context.fb_adr &= 0x0000007ff0000000;
			break;
		case 0x6:
			size = 512*1024*1024;
			context->device_context.fb_adr &= 0x0000007fe0000000;
			break;
		case 0x0:
			size = 128*1024*1024;
			context->device_context.fb_adr &= 0x0000007ff8000000;
			break;
		default:
			EMGD_ERROR_EXIT("Invalid Aperture Size");
			return -IGD_ERROR_NODEV;
			break;
		}
	}

	EMGD_DEBUG("MSAC=0x%x, GMADR Aperture Size=%dMB, fb_adr=0x%llx",tmp, size/(1024*1024), context->device_context.fb_adr);

	/* In SNB MMIO and GTT are together
	 * MMIO base = GTTMMADR
	 * GTTADR base = GTTMMADR + 2MB
	 */
	if(OS_PCI_READ_CONFIG_32(platform_context->igfx_dev.pcidev0,
			INTEL_OFFSET_VGA_MMADR,
			(void*)&context->device_context.mmadr)) {
		EMGD_ERROR_EXIT("Reading MMADR");
		return -IGD_ERROR_NODEV;
	}
	context->device_context.mmadr &= 0x00000000ffc00000;

	OS_PCI_READ_CONFIG_32(platform_context->igfx_dev.pcidev0,
		INTEL_OFFSET_VGA_MMADR+4,(void*)&tmp32);
	tmp &= 0x0000007f;

	context->device_context.mmadr |= ((unsigned long long)tmp32 << 32);

	context->device_context.virt_mmadr =
		OS_MAP_IO_TO_MEM_NOCACHE(
			context->device_context.mmadr,
			gttmmadr_size);
	EMGD_DEBUG("MMADR mapped %luKB @ (phys):0x%llx  (virt):%p",
		gttmmadr_size/2048,
		context->device_context.mmadr,
		context->device_context.virt_mmadr);

	/* GTT is directly after MMIO Space. */
	/* In Koheo, virt_gttadr is declared as "unsigned long *",
	 * not "unsigned char *" as in IEGD. Have to becareful when
	 * access GTT with virt_gttadr pointer, each increment of the pointer
	 * is equal to 4 bytes in address
	 */
	context->device_context.gttadr =
		(context->device_context.mmadr + 2*1024*1024);
	context->device_context.virt_gttadr =
		(unsigned long *)(context->device_context.virt_mmadr + 2*1024*1024);

	/* Number of Entries in the GTT (or size of the GTT in DWORDS).
	 * One Entry per page in the Aperture. */

	/* BIOS can allocate GTT size of 2MB, but not all the GTT are used
	 * For aperture size of 512MB, GTT size used is (512M/4k)*4 = 512k
	 * For aperture size of 256MB, GTT size used is (256M/4k)*4 = 256k
	 * gatt_pages = aperture_size/4k (=GTT entries)
	 */
	context->device_context.gatt_pages = (size/4096);

	EMGD_DEBUG("GTTADR mapped %luKB @ (virt):%p",
		gttmmadr_size/2048,
		context->device_context.virt_gttadr);
	EMGD_DEBUG("GTT actual size %ldKB entries",
		context->device_context.gatt_pages/1024);

	/* PCI Interrupt Line */
	if(OS_PCI_READ_CONFIG_8(platform_context->igfx_dev.pcidev0,
			PCI_INTERRUPT_LINE,
			(void*)&platform_context->igfx_dev.irq)) {
		platform_context->igfx_dev.irq = 0;
	}

	EMGD_TRACE_EXIT;
	return 0;
}

/*!
 * Get the # of pages used for video memory.  This does not use information from
 * the scratch register, since this is done later if it exists.
 * 
 * @param *context
 * @param *pages
 * 
 * @return -IGD_ERROR_INVAL on failure
 * @return 0 on success
 */
static int get_stolen_mem_snb(igd_context_t *context, unsigned long *pages)
{
	platform_context_snb_t *platform_context;
	os_pci_dev_t           vga_dev;
	unsigned short         gmch_ctl;
	unsigned long          stolen_mem; /* in bytes */
	int ret;

	EMGD_TRACE_ENTER;

	platform_context = (platform_context_snb_t *)context->platform_context;
	vga_dev          = platform_context->dramc_dev.pcidev;

	ret = OS_PCI_READ_CONFIG_16(vga_dev, PCI_GCC1_REG, &gmch_ctl);
	if (ret) {
		EMGD_ERROR_EXIT("PCI read of GMCH Ctl");
		return -IGD_ERROR_INVAL;
	}

	EMGD_DEBUG("PCI_GCC_REG = 0x%x",gmch_ctl);
	
	stolen_mem = ((gmch_ctl & 0xf8) >> 3) * 32*1024*1024;

	if(stolen_mem > 512*1024*1024) {
		EMGD_ERROR_EXIT("Unknown Stolen Memory Size");
		return -IGD_ERROR_INVAL;
	}
	
	EMGD_DEBUG("Total stolen memory read from PCI_GCC_REG = %ldMB", stolen_mem/(1024*1024));

	if (stolen_mem) {
		/* Subtract size of the GTT which is
		 * (number of entries in DWORDS) * 4 to get it into bytes */
		/* stolen_mem -= context->device_context.gatt_pages*4; */
		/* Subtract off 1 page for the scratch page */
		stolen_mem -= 4*1024;
	}

	/* Convert to the # of pages available for stolen memory */
	*pages = stolen_mem / 4096;

	EMGD_DEBUG("Adjsuted stolen memory: 0x%lx pages", *pages);

	EMGD_TRACE_EXIT;
	return 0;
}

/*!
 * Program the GTT.
 * 
 * @param context
 * 
 * @return gtt_init_ret_code
 * @return 0
 */
int gtt_init_snb(igd_context_t *context)
{
	/*
	 * Program the GTT.
	 * The GTT is programmed as follows when NOT using A000 for storing
	 * Dynamic Data:
	 *                     Generic case             256MB RAM with
	 *                                            128MB Aperture and
	 *                                            8MB Stolen memory
	 *                                                 Example
	 *                 --------------------     --------------------
	 *     ----------  Top of RAM (TR)                256MB
	 *     |  GTT   |
	 * S M |        |
	 * t e |--------|  TR-sizeof(GTT)                 256MB-128KB
	 * o m |Scrch pg|  4KB for scratch page
	 * o o |--------|                                 256MB-132KB
	 * e r |        |
	 * n y | Video  |
	 *     |  Mem   |
	 *     |--------|  TR-sizeof(stolen_mem)          256MB-8MB
	 *     |        |
	 *     | System |
	 *     |  Mem   |
	 *     |        |
	 *     ----------  00000000                       0
	 *
	 *
	 * The GTT is programmed as follows when using A000 for storing
	 * Dynamic Data:
	 *                   Generic case         256MB RAM with 128MB
	 *                                          Aperture Example
	 *               --------------------     --------------------
	 *     ----------  Top of RAM (TR)                256MB
	 *     |  GTT   |
	 * S   |        |
	 * t   |--------|  TR-sizeof(GTT)                 256MB-128KB
	 * o M |Scrch pg|  4KB for scratch page
	 * l e |--------|                                 256MB-132KB
	 * e m |Dyn A000|  60KB for Dynamic A000 Data
	 * n o |--------|                                 256MB-192KB
	 *   r |        |
	 *   y | Video  |
	 *     |  Mem   |
	 *     |--------|  TR-sizeof(stolen_mem)          256MB-8MB
	 *     |        |
	 *     | System |
	 *     |  Mem   |
	 *     |        |
	 *     ----------  00000000                       0
	 *
	 * Program the GTT for CONFIG_Q35.
	 *
	 * The GTT is programmed as follows when using A000 for storing
	 * Dynamic Data:
	 *
	 *   sizeof(GTT) = 1MB with VT
	 *   sizeof(GTT) = 2MB without VT
	 *
	 *                   Generic case            256MB RAM with
	 *                                        128 MB Stolen Memory
	 *               --------------------     --------------------
	 *     ----------  Top of RAM (TR)                256MB
	 * S   |Scrch pg|  
	 * t   |--------|  4KB for scratch page           256MB-4k
	 * o M |Dyn A000|  
	 * l e |--------|  60KB for Dynamic A000 Data     256MB-64k
	 * e m | Video  | 
	 * n o |  Mem   |
	 *   r |        |
	 *   y |--------|  TR-sizeof(stolen_mem)          256MB-128MB
	 *     | GTT Mem|  
	 *     | Stolen |
	 *     |--------|  TR-sizeof(GTT_stolen_mem)      256MB-128MB-1MB
	 *     |        |
	 *     | System |
	 *     |  Mem   |
	 *     |        |
	 *     ----------  00000000                       0
	 */
	int gtt_init_ret_code;
	os_pci_dev_t vga_dev;
	platform_context_snb_t *platform_context;
	unsigned long phys_stolen_mem, /*phys_page_table,*/ page_table_entry;
	uint32_t pci_read_long;
	unsigned char *mmio;
	unsigned long *virt_gttadr;
	struct page *scratch_page;
	unsigned long pte;
	
	EMGD_TRACE_ENTER;

	platform_context = (platform_context_snb_t *)context->platform_context;
	vga_dev = platform_context->igfx_dev.pcidev0;
	mmio = context->device_context.virt_mmadr;
	virt_gttadr = context->device_context.virt_gttadr;

	/* Enable the PCI Device */
	gtt_init_ret_code = OS_ENABLE_PCI(vga_dev);
	if(gtt_init_ret_code) {
		return gtt_init_ret_code;
	}

	/* Create a scratch page using alloc_page for compatibility with Koheo */
	scratch_page = alloc_page(GFP_DMA32 | __GFP_ZERO);
	context->device_context.scratch_page = scratch_page;

	/* In SNB, GTT is not at above graphics stolen memmory,
	 * stolen_pages = reserved_mem found earlier
	 * stolen_pages is initialized here for compatibility with Koheo */
	context->device_context.stolen_pages = context->device_context.reserved_mem;

	/* Get Stolen Memory Base Address */
	OS_PCI_READ_CONFIG_32(vga_dev, PCI_BSM_REG, &pci_read_long);
	context->device_context.reserved_mem_base_addr =
		pci_read_long & 0xFFF00000;

	EMGD_DEBUG("PCI_BSM_REG = %x", pci_read_long);

	/* Physical address of the beginning of video memory */
	phys_stolen_mem = context->device_context.reserved_mem_base_addr;

/* PGTBL_CTL cannot be used for Gen6 onwards, physical GTT is allocated by BIOS */
#if 0
	phys_page_table = phys_stolen_mem - 2*1024*1024;

	/* Set the Base of the GTT page table so each entry can be set.
	 * Also enable the GTT */
	EMGD_WRITE32(phys_page_table|1, mmio + PGTBL_CTL);
#endif
	EMGD_WRITE32(context->device_context.reserved_mem, mmio + 0x71418);

	/* Page table entry for scratch page */
	pte = page_to_phys(scratch_page) | 1;
	EMGD_DEBUG("Scratch page phys addr = 0x%llx",
		(unsigned long long)page_to_phys(scratch_page));

#if 0
	/* Program the GTT */
	for(page_table_entry = 0;
		page_table_entry < context->device_context.reserved_mem;
		page_table_entry++) {
		/* Enable */
		EMGD_WRITE32(phys_stolen_mem|1, virt_gttadr + page_table_entry);
		phys_stolen_mem += 4096;
	}

	/* Program the remainder of the GTT to the scratch page. */
	for(page_table_entry = context->device_context.reserved_mem;
		page_table_entry < (context->device_context.gatt_pages);
		page_table_entry++) {
		/* Enable */
		EMGD_WRITE32(pte, virt_gttadr + page_table_entry);
	}
#else
	/* For SNB debug purpose, program the whole GTT with scratch page */
	for(page_table_entry = 0;
		page_table_entry < (context->device_context.gatt_pages);
		page_table_entry++) {
		/* Enable */
		EMGD_WRITE32(pte, virt_gttadr + page_table_entry);
	}
#endif
	EMGD_TRACE_EXIT;
	return 0;
}


